// $Id: bitmap.c,v 1.1 1997/03/25 10:18:43 yotam Exp $

#include <string.h>
#include <BitSet.h>
#include "bitmap.h"
#include "bmout.h"
#include "gcd.h"
#include "debug.h"

////////////////////////////////////////////////////////////////////////
Bitmap::Bitmap() :
   _w(0), _h(0), _bitSet(0)
{} // Bitmap


////////////////////////////////////////////////////////////////////////
Bitmap::Bitmap(const Bitmap& sample) :
   _w(sample._w), 
   _h(sample._h), 
   _bitSet(0)
{
   if (_h) 
   {
      _bitSet = new BitSet[_h];
   }
   int  i;
   for (i = _h;  i--;)
   {
      _bitSet[i] = sample._bitSet[i];
   }
} // Bitmap


////////////////////////////////////////////////////////////////////////
Bitmap::Bitmap(const Bitmap &bm, unsigned int n, unsigned int d) :
   _w(0), 
   _h(0), 
   _bitSet(0) 
{
   if (d)
   {
      unsigned int  div = gcd(n, d);
      n /= div;
      d /= div;
      _w = (bm._w * n + d - 1) / d;
      _h = (bm._h * n + d - 1) / d;
      _bitSet = new BitSet[_h];
      magnify(bm, n, d);
   }
} // Bitmap


////////////////////////////////////////////////////////////////////////
Bitmap::Bitmap
(
   unsigned int           w, 
   unsigned int           h,        
   const BitSet * const * bitSetPArray
) :
   _w(w), 
   _h(h), 
   _bitSet(0)
{
   if (h)
   {
      _bitSet = new BitSet[h];
      int  i;
      for (i = h;  i--;)
      {
         _bitSet[i] = *(bitSetPArray[i]);
      }
   }
} // Bitmap


////////////////////////////////////////////////////////////////////////
Bitmap::~Bitmap()
{
   if (_bitSet)
   {
      delete [] _bitSet;
   }
} // ~Bitmap


////////////////////////////////////////////////////////////////////////
Bitmap & 
Bitmap::operator = (const Bitmap &rhs)
{
   if (this != &rhs)
   {
      if (_h < rhs._h)
      {
         if (_bitSet)
         {
            delete [] _bitSet;
         }
         _bitSet = new BitSet[rhs._h];
      }
      _w = rhs._w;
      _h = rhs._h;
      int  i;
      for (i = _h;  i--;)
      {
         _bitSet[i] = rhs._bitSet[i];
      }
   }
   return(*this);
} // opearotor =


////////////////////////////////////////////////////////////////////////
void
Bitmap::clear()
{
   if (_bitSet)
   {
      delete [] _bitSet;
      _bitSet = 0;
   }
   _w = _h = 0;
} // clear


////////////////////////////////////////////////////////////////////////
unsigned int
Bitmap::value(unsigned int x, unsigned int y) const
{
   unsigned int  rc = 0;
   if (y < _h)
   {
      rc = _bitSet[y][x];
   }
   return(rc);
} // value;


////////////////////////////////////////////////////////////////////////
void
Bitmap::set
(
   unsigned int x, 
   unsigned int y,
   unsigned int value
)
{
   bool inWH = ((x < _w) && (y < _h));
   if (inWH || value)
   {
      if (y >= _h)
      {
         realloc(y);
      }
      BitSet  &bs = _bitSet[y];
      if (value)
      {
         bs.set(x);
      }
      else
      {
         bs.clear(x);
      }
   }
} // set


////////////////////////////////////////////////////////////////////////
void
Bitmap::bound(unsigned int /* w */, unsigned int h)
{
   realloc(h);
} // bound


////////////////////////////////////////////////////////////////////////
void
Bitmap::realloc(unsigned int h)
{
   if (_h < h)
   {
      BitSet  *newBS = new BitSet[h];
      if (_bitSet)
      {
         int  i;
         for (i = _h;  i--;)
         {
            newBS[i] = _bitSet[i];
         }
         delete [] _bitSet;
      }
      _bitSet = newBS;
      _h = h;
   }
} // bound


////////////////////////////////////////////////////////////////////////
// Given the map:  x |-->  (/dn) * x 
// Now look at the mapping of the segment [x, x+1).
// Output:
//   [beginFull, endFull) = the integral segment that is completely covered.
//   preBegin / d         = the fraction covered of [beginFull-1, beginFull)
//   postBegin / d        = the fraction covered of [endFull, endFull+1)
static void
getCoverDomains
(
   // output
   unsigned int&  preBegin,
   unsigned int&  beginFull,
   unsigned int&  endFull,
   unsigned int&  postEnd,
   // input
   unsigned int   n,
   unsigned int   d,
   unsigned int   x
)
{
   unsigned int  nx  = n * x;
   unsigned int  nx1 = nx + n;
   beginFull         = (nx + d - 1) / d;
   endFull           = nx1 / d;
   preBegin          = d * beginFull - nx;
   postEnd           = nx1 % d;
   DASSERT(preBegin == (d - (nx % d)) % d);
   DASSERT(postEnd  == nx1 - d * endFull);
} // getCoverDomains


////////////////////////////////////////////////////////////////////////
// Assuming size and allocation well set
void 
Bitmap::magnify(const Bitmap &bm, unsigned n, unsigned d)
{
   unsigned  wh = _w * _h;
   unsigned  *darkMatrix = new unsigned int[wh];
   unsigned  darkSum = 0;
   int       i, j;
   (void)memset(darkMatrix, 0, wh * sizeof(unsigned));

   for (i = bm.w();  i--;)
   {
      unsigned int    xpb, xb, xe, xpe;
      getCoverDomains(xpb, xb, xe, xpe, n, d, i);
      for (j = bm.h();  j--;)
      {
         if (bm.value(i, j))
         {
            unsigned int    ypb, yb, ye, ype;
            getCoverDomains(ypb, yb, ye, ype, n, d, j);

            // 3x3=9 (dx,dy) cases:  {xpb, d, xpe} X {ypb, d, ype}
            int  x;  unsigned int  dx;
            for (x = xb - 1, dx = xpb;  x <= (int)xe;  x++)
            {
               if (dx)
               {

                  int  y;  unsigned int  dy;
                  for (y = yb - 1, dy = ypb;  y <= (int)ye;  y++)
                  {
                     if (dy)
                     {
                        unsigned             m  = dx * dy;
                        darkSum                += m;
                        darkMatrix[_w * y + x] += m;
                     }
                     dy = d;
                     if (y+1 == (int)ye)
                     {
                        dy = ype;
                     }
                  }

               }
               dx = d;
               if (x+1 == (int)xe)
               {
                  dx = xpe;
               }
            }

         }
      }
   }

#if 0
   // now determine by the darkMatrix which bits to be set
   unsigned threshold = d * d / 2;
   int  x, y;
   for (x = _w;  x--;)
   {
      for (y = _h;  y--;)
      {
         if (darkMatrix[_w * y + x] > threshold)
         {
            set(x, y, 1);
         }
      }
   }
   // May be ebtter to sort index to darkMatrix, and blacken until
   // we are fine.
#endif

   // Rather clumsy. simply lower the threshold until we are done.
   unsigned  d2 = d * d, threshold, nextThreshold = d2 / 2 + 1, darkness;
   do
   {
      int       x, y;
      threshold = nextThreshold;
      nextThreshold = 0;
      darkness = 0;
      for (x = _w;  x--;)
      {
         for (y = _h;  y--;)
         {
            unsigned  pixDark = darkMatrix[_w * y + x];
            if (pixDark >= threshold)
            {
               set(x, y, 1);
               darkness += d2;
            }
            else if (nextThreshold < pixDark)
            {
               nextThreshold = pixDark;
            }
         }
      }
   } while ((darkness < darkSum) && nextThreshold);
   delete [] darkMatrix;
} // magnify


////////////////////////////////////////////////////////////////////////
ostream&
operator <<(ostream& os, const Bitmap& bm)
{
   typedef const BitSet  *pBitSet;
   unsigned int          h = bm.h();
   pBitSet               *pa = new pBitSet[h];
   const BitSet          *a  = bm.bitSetArray();
   int  i;
   for (i = h;  i--;)
   {
      pa[i] = &a[i];
   }
   bitmapOutput(os, pa, bm.w(), h);
   return(os);
}


////////////////////////////////////////////////////////////////////////
BitmapRectIterator::BitmapRectIterator(const Bitmap *pbm) : 
   _x(0), _y(0),  _pbm(pbm)
{
   _covered.bound(pbm->w(), pbm->h());
} // BitmapRectIterator


////////////////////////////////////////////////////////////////////////
BitmapRectIterator::~BitmapRectIterator()
{
} // ~BitmapRectIterator


////////////////////////////////////////////////////////////////////////
static void
fold(unsigned int& x, unsigned int& y, unsigned int bmw/*, unsigned int bmh*/)
{
   if (x == bmw)
   {
#if 0
      if (++y < bmh)
      {
         x = 0;
      }
#endif
      x = 0;
      y++;
   }
} // fold


////////////////////////////////////////////////////////////////////////
bool
BitmapRectIterator::nextRect
(
   unsigned int& x, 
   unsigned int& y, 
   unsigned int& w, 
   unsigned int& h
)
{
   bool           got = false;
   unsigned int   bmw = _pbm->w();
   unsigned int   bmh = _pbm->h();

   // Get to the next uncovered black pixel
   while ((_x < bmw) && (_y < bmh) && !got)
   {
      got = _pbm->value(_x, _y) && !_covered.value(_x, _y);
      if (!got)
      {
         _x++;
         fold(_x, _y, bmw /*, bmh */);
      }
   }

   if (got)
   {
      x = _x;  y = _y;

      // Get maximal width. Note that by the algorithm, 
      // For all set bits scanned in the segment are not yet covered.
      while ((++_x < bmw) && (_pbm->value(_x, _y) != 0));
      unsigned int  xPost = _x;
      w = _x - x;  
      fold(_x, _y, bmw /* , bmh */);

      // Get maximal height. 
      // Heuristics: We will grow only if the the above segment exactly match.
      bool          grow = true;
      unsigned int  ySucc = y + 1;
      while (grow && (ySucc < bmh))
      {
         // see that outside boundary is also cleared at the ySucc row.
         grow = ((x == 0)       || (_pbm->value(x - 1, ySucc) == 0)) &&
                ((xPost == bmw) || (_pbm->value(xPost, ySucc) == 0));
         unsigned  xi;
         for (xi = x;  grow && (xi < xPost);  xi++)
         {
            grow = (_pbm->value(xi, ySucc) != 0);
         }
         if (grow)
         {
            for (xi = x;  xi < xPost;  xi++)
            {
               _covered.set(xi, ySucc);
            }
            ySucc++;
         }
      }
      h = ySucc - y;
   }

   return(got);
} // nextRect
