// $Id: lspath.c,v 1.7 1997/07/25 05:44:30 yotam Exp yotam $
//
// lspath -- Interface for a TCL function.
//           Result is a filtered list of path-file names of given path.
// Usage:
//          lspath [-base <basename>]
//                 [-regexp {base|full} <exp>]
//                 [-proccond {base|full} <procname+leadargs>]
//                 <path>
//
// See further documentation in  lspath.[tex -> dvi -> ps]

#if DEBUG0
#define NSUBEXP  20

typedef struct regexp {
	char* startp[NSUBEXP];
	char* endp[NSUBEXP];
	char regstart;		/* Internal use only. */
	char reganch;		/* Internal use only. */
	char *regmust;		/* Internal use only. */
	int regmlen;		/* Internal use only. */
	char program[1];	/* Unwarranted chumminess with compiler. */
} regexp;
#endif

#include <string>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#if defined(S_IFLNK) && defined(_S_IFMT) && !defined(S_ISLNK)
   // POSIX, and DEC ULTRIX, lack  S_ISLNK() 
#  define S_ISLNK(mode) (((mode) & _S_IFMT) == S_IFLNK)
#endif

#include <iostream.h>
#include <strstream.h>

#include <tcl.h>

#include "pushdir.h"
#include "debug.h"



class LspathBlock
{
 public:
   enum {MaxFileName = 256};
   Tcl_Interp*  interp;
 
   char*       path;           // working copy
   string      node;           // current traversed node. directory/file
   ostrstream  sresult;        // working result

   LspathBlock();
   virtual ~LspathBlock();

   int   parseArgs(int argc, const char*  const*  const argv);
   int   dirsp(DIR* tdir, char* taildir);
   void  setResult();

 private:

   LspathBlock(const LspathBlock&);   // declared but undefined!
   LspathBlock& operator=(const LspathBlock&);   // declared but undefined!
   
   const char*  baseTarget;    // The callback predicate command
   char*        condCmd;       // The callback predicate command
   char*        condFnPtr;     // Into condCmd
   int          condBaseFlag;  // predicate acts on base name only.

   char*       regexp;
   Tcl_RegExp  compRegexp;
   int         regexpBaseFlag; // regexp filter acts ib base name only

   int tree(DIR* tdir, char* taildir);
   int AddFiles(DIR* tdir);
   int AddFile(const char* basename);
}; // LspathBlock


// Assuming very specific behaviour of the pointed String
class ChildDir
{
 public: 
   ChildDir(string &node, const char* childName);
   ~ChildDir();
   void       pop();
   DIR*        dir() {return _dir;}; 
 private:
   string &   wnode;    
   const int  currDirLen;
   DIR*        _dir; 
   int        symbolic; 
}; // ChildDir


////////////////////////////////////////////////////////////////////////
static int
newstrcpy(Tcl_Interp* interp, char* &ns, const char* s)
{
   int  rc = TCL_OK;
   int  len = strlen(s);
   ns   = new char[len + 1];
   if (ns == 0)
   {
      rc = TCL_ERROR;
      (void)sprintf(interp->result, "No memory to copy %d size string", len);
   }
   else
   {
      (void)strcpy(ns, s);
   }
   return(rc);
} // newstrcpy


////////////////////////////////////////////////////////////////////////
static void
safeDelete(char*  &p)
{
   if (p)
   {
      delete [] p;
      p = 0;
   }
} // safeDelete


////////////////////////////////////////////////////////////////////////
LspathBlock::LspathBlock() :
   interp(0),
   path(0),
   baseTarget(0),
   condCmd(0),
   regexp(0),
   compRegexp(0)
{
} // LspathBlock::LspathBlock


////////////////////////////////////////////////////////////////////////
LspathBlock::~LspathBlock()
{
   safeDelete(path);
   safeDelete(condCmd);
   safeDelete(regexp);
   // compRegexp is cached internally by TCL we must not free it.
   // if (compRegexp) free(compRegexp);
} // LspathBlock::~LspathBlock


////////////////////////////////////////////////////////////////////////
ChildDir::ChildDir(string &node, const char* childName):
   wnode(node),
   currDirLen(node.length()),
   _dir(0)
{
   DPR("ChildDir(" << node << ", " << childName << ")" << endl);
   struct stat  statBuf;        
   if ((stat(childName, &statBuf) == 0) && S_ISDIR(statBuf.st_mode))
   {
      symbolic = S_ISLNK(statBuf.st_mode);
      if (chdir(childName) == 0)
      {
         _dir = opendir(".");
         if (_dir)
         {
            node += '/';
            node += childName;
         }
      }
   }
} // ChildDir::ChildDir


////////////////////////////////////////////////////////////////////////
ChildDir::~ChildDir()
{
   DPR("~ChildDir: _dir=" << _dir << ", wnode=" << wnode << endl);
   pop();
} // ChildDir::~ChildDir


////////////////////////////////////////////////////////////////////////
void
ChildDir::pop()
{
   DPR("~ChildDir: _dir=" << _dir << ", wnode=" << wnode << endl);
   if (_dir)
   {
      closedir(_dir);
      wnode.erase(currDirLen);
      const char* updir = (symbolic ? wnode.c_str() : "..");
      if (chdir(updir) != 0) 
      {
         cerr << "Failed to chdir to " << updir << endl;
      }
      _dir = 0;
   }
} // ChildDir::~ChildDir


////////////////////////////////////////////////////////////////////////
static void
freebyDelete(char* p)
{
   delete [] p;
} // freebyDelete


////////////////////////////////////////////////////////////////////////
void
LspathBlock::setResult()
{
   sresult << ends;
   int len = sresult.pcount();
   if (len >= 2) 
   {
      const char*  s = sresult.str(); // imply freeze
      int          l = strlen(s);   DASSERT(l + 1 == len);
      char*        rs = new char[l + 1];
      strcpy(rs, s);
      sresult.freeze(0);
      rs[len - 1] = '\0'; // trim last space
      DPRNL("rs=" << rs << '.');
      Tcl_SetResult(interp, rs, freebyDelete);
   }
   else
   {
      interp->result[0] = '\0';
   }
} // LspathBlock::setResult


////////////////////////////////////////////////////////////////////////
int
LspathBlock::AddFile(const char* basename)
{
   DPR("AddFile(basename=" << basename << ")" << endl);
   const char* filename = node.c_str();
   int  add = (strchr(filename, ' ') == 0), // ignore filenames with space
        rc = TCL_OK;
   if (add && baseTarget)
   {
      add = (strcmp(basename, baseTarget) == 0);
   }
   if (add && regexp)
   {
      char* searced = (char *)(regexpBaseFlag ? basename : filename);
      int  rex = Tcl_RegExpExec(interp, compRegexp, searced, searced);
      add = (rex == 1);
      if (rex == -1)
      {
         cerr << "Error: " << interp->result << endl;
         DASSERT(false);
      }
                            
   }
   if (add && condCmd)
   {
      (void)strncpy(condFnPtr,
                    (condBaseFlag ? basename : filename),
                    LspathBlock::MaxFileName);
      rc = Tcl_Eval(interp, condCmd);
      add = 0;
      if (rc == TCL_OK)
      {
         rc = Tcl_ExprBoolean(interp, interp->result, &add);
      }
   }
   if (add)
   {
      sresult << filename << ' ';
   }
   return(rc);
} // LspathBlock::AddFile


////////////////////////////////////////////////////////////////////////
inline static int
dotORdotdot(const char* node)
{
   int  rc = // inner loop - let's be efficient
             ((node[0] == '.') &&
                ((node[1] == '\0') ||
                 ((node[1] == '.') && (node[2] == '\0'))
                )
             );
   return(rc);
} // dotORdotdot


////////////////////////////////////////////////////////////////////////
int
LspathBlock::AddFiles(DIR* dir)
{
   DPR("AddFiles, node=" << node << endl);
   int  rc = TCL_OK;

   node += '/';
   int  dlen = node.length();
   struct dirent*  dent;
   while ((dent = readdir(dir)) && (rc == TCL_OK))
   {
      const char* d_name = dent->d_name;
      if (!dotORdotdot(d_name))
      {
         node += d_name;
         rc = AddFile(d_name);
         node.erase(dlen);
      }
   }

   node.erase(dlen - 1); // trim last '/'

   return(rc);
} // LspathBlock::AddFiles


////////////////////////////////////////////////////////////////////////
int
LspathBlock::dirsp(DIR* tdir, char* taildir)
{
   // Any "//" in taildir is interpreted as spanning the directory subtree.

   DPR("dirsp(" << taildir << "), node=" << node << endl);
   int  rc = TCL_OK;

   int  taildirLength = strlen(taildir);
   if (taildirLength == 0)
   {
      if (tdir)
      {
         rc = AddFiles(tdir);
      }
   }
   else
   {
      char*  dblSlash = strstr(taildir, "//");
      int pathLen = node.length();
      if (dblSlash)
      {
         *dblSlash = '\0'; // temporary trim
      }
      node += taildir; 
      PushDir  pushdir(node.c_str()); // will implicitly pop
      DIR*      dir = pushdir.dir();
      if (dir)
      {
         if (dblSlash)
         {
            rc = tree(dir, dblSlash + 1);  // single / may be handy
         }
         else
         {
            rc = AddFiles(dir);
         }
         pushdir.pop();
      }
      node.erase(pathLen);
      if (dblSlash)
      {
         *dblSlash = '/'; // restore  
      }
   }
   return(rc);
} // LspathBlock::dirsp


////////////////////////////////////////////////////////////////////////
static bool
hasSubDirs(const char *snode)
{
   bool         has = false;
   struct stat  statNode;
   if (stat(snode, &statNode) == 0)
   {
      has = (statNode.st_nlink > 2); // Kpathsea trick
   }
   return(has);
} // hasSubDirs


////////////////////////////////////////////////////////////////////////
int
LspathBlock::tree(DIR* tdir, char* taildir)
{
   DPR("tree(" << taildir << "), node=" << node << endl);
   if (strlen(taildir) == 1) // == "/"
   {
      taildir = "";
   }
   int  rc = dirsp(tdir, taildir);
   if ((rc == TCL_OK) && hasSubDirs(node.c_str()))
   {
      struct dirent*  dent;
      rewinddir(tdir);
      while ((dent = readdir(tdir)) && (rc == TCL_OK))
      {
         const char* d_name = dent->d_name;
         if (!dotORdotdot(d_name))
         {
            ChildDir  childDir(node, d_name); // push and implicit pop
            DIR*  dir = childDir.dir();
            if (dir)
            {
               rc = tree(dir, taildir);
               childDir.pop();
            }
         }
      }
   }
   return(rc);
} // LspathBlock::tree


////////////////////////////////////////////////////////////////////////
static int
baseORfull(Tcl_Interp* interp, int &baseFlag, const char* opt)
{
   int  rc = TCL_OK;

   if (strcmp(opt, "base") == 0)
   {
      baseFlag = 1;
   }
   else if (strcmp(opt, "full") == 0)
   {
      baseFlag = 0;
   }
   else
   {
      rc = TCL_ERROR;
      (void)sprintf(interp->result,
                    "'base' or 'full' expected instead of %s", opt);
   }
   return(rc);
} // baseORfull


////////////////////////////////////////////////////////////////////////
int
LspathBlock::parseArgs(int argc, const char*  const*  const argv)
{
   int  rc = TCL_OK, ai, len;

   DASSERT(strcmp(argv[0], "lspath") == 0);
   for (ai = 1;  (rc == TCL_OK) && ai < argc - 1;)
   {
      if (strcmp(argv[ai], "-base") == 0)
      {
         if (ai + 1 <= argc + 1)
         {
            baseTarget = argv[ai + 1];
            ai += 2;
         }
         else
         {
            rc = TCL_ERROR;
            (void)strcpy(interp->result, "-base: missing arguments");
         }
      }
      else if (strcmp(argv[ai], "-regexp") == 0)
      {
         if (ai + 3 <= argc - 1)
         {
            rc = baseORfull(interp, regexpBaseFlag, argv[ai + 1]);
            if (rc == TCL_OK)
            {
               rc = newstrcpy(interp, regexp, argv[ai + 2]);
               compRegexp = Tcl_RegExpCompile(interp, regexp);
               if (compRegexp == 0)
               {
                  rc = TCL_ERROR;
               }
            }
         }
         else
         {
            rc = TCL_ERROR;
            (void)strcpy(interp->result, "-regexp: missing arguments");
         }
         ai += 3;
      }
      else
      if (strcmp(argv[ai], "-proccond") == 0)
      {
         if (ai + 3 <= argc - 1)
         {
            rc = baseORfull(interp, condBaseFlag, argv[ai + 1]);
            if (rc == TCL_OK)
            {
               // copy the predicate command leaving a place for file name
               len = strlen(argv[ai + 2]);
               int  size = len + 1 + LspathBlock::MaxFileName + 1;
               condCmd = new char[size];
               if (condCmd == 0)
               {
                  (void)strcpy(interp->result, "No Memory for path");
                  rc = TCL_ERROR;
               }
               else
               {
                  (void)strcpy(condCmd, argv[ai + 2]);
                  (void)strcpy(condCmd + len, " ");
                  condFnPtr = condCmd + len + 1;
               }
            }
            ai += 3;
         }
         else
         {
            rc = TCL_ERROR;
            (void)strcpy(interp->result, "-proccond: missing arguments");
         }
      }
      else
      {
         (void)sprintf(interp->result, "Unknown option %s", argv[ai]);
         rc = TCL_ERROR;
      }
   }
   if (rc == TCL_OK)
   {
      if (ai != argc - 1)
      {
         rc = TCL_ERROR;
         (void)strcpy(interp->result, "Usage: \n"
                      "  lspath  ?-base <basename>?\n"
                      "          ?-regexp {base|full} <exp>?\n"
                      "          ?-proccond {base|full} <procname+leadargs>?\n"
                      "          <path>\n");
      }
      else
      {
         // copy the path, it may be a read only constant
         len = strlen(argv[ai]);
         path = new char[len + 1];
         if (path == 0)
         {
            (void)strcpy(interp->result, "No Memory for path");
            rc = TCL_ERROR;
         }
         else
         {
            (void)strcpy(path, argv[ai]);
         }
      }
   }
   return(rc);
} // LspathBlock::parseArgs


////////////////////////////////////////////////////////////////////////
int
lspath(ClientData, Tcl_Interp* interp, int argc, char** argv)
{
   DPR("lspath>>" << endl);
   LspathBlock  lspathBlock;
   lspathBlock.interp = interp;

   int  rc = lspathBlock.parseArgs(argc, argv);

   char*  currp = lspathBlock.path;
   while ((rc == TCL_OK) && *currp)
   {
      char* pcol = strchr(currp, ':');
      if (pcol)
      {
         *pcol = '\0';
      }
      lspathBlock.node = "";
      rc = lspathBlock.dirsp(0, currp);
      currp = (pcol ? pcol + 1 : "");  // skip ex-colon or end
   }
   if (rc == TCL_OK)
   {
      lspathBlock.setResult();
   }
   DPR("lspath<<" << endl);
   return(rc);
} // lspath
