#include <string>
#include <string.h>
#include <iostream.h>
#include <algo.h>
#include <vector.h>
#include "debug.h"


class FontFile
{
 public:
   FontFile(const char *fn = 0);

   class LessBase
   {
    public:
      bool operator()(const FontFile& ff0, const FontFile& ff1)
      {  return(diffBase(ff0, ff1) < 0);      }
   }; // LessBase

   class EqBase
   {
    public:
      bool operator()(const FontFile& ff0, const FontFile& ff1)
      {  return(diffBase(ff0, ff1) == 0);      }
   }; // EqBase


   class LessName
   {
    public:
      bool operator()(const FontFile& ff0, const FontFile& ff1)
      {  
         int  cmp = strcmp(ff0.name.c_str(), ff1.name.c_str());
         return(cmp < 0);
      }
   }; // LessName

   static int diffBase(const FontFile& ff0, const FontFile& ff1)
   {  
      int  cmp = strcmp(ff0.base, ff1.base);
      return(cmp);
   }

   const char*  filename;
   const char*  base;     // ptr into filename
   string       name;     // base with suffix trimmed

}; // FontFile
typedef FontFile*  FontFileP;

////////////////////////////////////////////////////////////////////////
FontFile::FontFile(const char* fn):
   filename(fn),
   base(fn),
   name("")
{
   if (filename)
   {
      const char*  lastSlash = strrchr(filename, '/');
      if (lastSlash)
      {
         base = lastSlash + 1;
      }
      const char*  dot = strrchr(base, '.');
      name = (dot ? string(base, dot - base) : string(base));
   }
} // FontFile::FontFile


////////////////////////////////////////////////////////////////////////
// argv[argc == nTfm + nPk] is an array of:
//    nTfm  .tfm files
//    nPk   .<nnn>pk files
// matchFonts writes on os (in TCL syntax) list of 3 lists:
// {{fontList} {tfmList} {pkList}}
// All three list are of the same length and matched by index.
//
// Example:
//     Say nTfm=2 nPk=3 argc=5 
//     argv[]= {
//                /usr/lib/tex/fonts/tfm/cmr10.tfm
//                /usr/lib/tex/fonts/tfm/cmr8.tfm
//                /usr/lib/tex/fonts/pk/cmr10.329pk
//                /usr/lib/tex/fonts/pk/cmr9.300pk
//                /usr/lib/tex/fonts/pk/cmr10.300pk
//             }
// Then the output will be (without end of lines)
// { {cmr10.300 cmr10.329}
//   {/usr/lib/tex/fonts/pk/cmr10.300pk /usr/lib/tex/fonts/pk/cmr10.329pk)
//   {/usr/lib/tex/fonts/tfm/cmr10.tfm /usr/lib/tex/fonts/tfm/cmr10.tfm} }
//
bool
matchFonts
(
   ostream&            os,
   int                 nTfm,
   int                 nPk,
   int                 DCODE(argc),
   const char* const*  argv
)
{
   DASSERT(nTfm + nPk == argc);
   bool  ok = true;
   int  i;
#if 0 && DEBUG && !defined(MAIN)
   cout << endl << "matchFonts " << nTfm << ' ' << nPk << ' ' << argc << endl;
   for (i = 0;  i < argc;  i++)
   {
      cout << argv[i] << endl;
   }
   cout << endl << endl;
#endif
   const char* const*  tfmFileNames = argv;
   const char* const*  pkFileNames = argv + nTfm;

   FontFileP  tfmFFs = new FontFile[nTfm];
   for (i = nTfm;  i--;  tfmFFs[i] = FontFile(tfmFileNames[i]));

   FontFileP  pkFFs = new FontFile[nPk];
   for (i = nPk;  i--;  pkFFs[i] = FontFile(pkFileNames[i]));

   // uniquley sort
   stable_sort(tfmFFs, tfmFFs + nTfm, FontFile::LessBase());
   FontFileP  endTfm = unique(tfmFFs, tfmFFs + nTfm, FontFile::EqBase());
   stable_sort(pkFFs, pkFFs + nPk, FontFile::LessBase());
   FontFileP  endPk = unique(pkFFs, pkFFs + nPk, FontFile::EqBase());

   // 1st pass just to get the indices to pkFFs and tfmFFs
   typedef int*  intP;
   intP  ind = new int[2 * (endPk - pkFFs)]; // good bound
   intP  pkIndBegin = ind, pkIndEnd = pkIndBegin,
         tfmIndBegin = ind + (endPk - pkFFs), tfmIndEnd = tfmIndBegin;

   for (FontFileP pkIter = pkFFs, begTfm = tfmFFs;  
        (pkIter != endPk) && (begTfm != endTfm);  pkIter++)
   {
      pair<FontFile*, FontFile*>   
         eqRng = equal_range(begTfm, endTfm, *pkIter, FontFile::LessName());
      if (eqRng.first != eqRng.second)
      {
         DASSERT(eqRng.first + 1 == eqRng.second);
         *pkIndEnd++ = (pkIter - pkFFs);
         *tfmIndEnd++ = (eqRng.first - tfmFFs);
      }
      begTfm = eqRng.first;
   }
   DASSERT((pkIndEnd-pkIndBegin) == (tfmIndEnd-tfmIndBegin))
   
   os << "{ {";  // font list:  foo.300pk -> foo.300  
   for (intP  ip = pkIndBegin;  ip != pkIndEnd;  ip++)
   {
      const char* base = pkFFs[*ip].base;
      int l = strlen(base);   DASSERT(l>2 && (strcmp(base +l-2, "pk") == 0));
      os.write(base, l - 2) << ' ';
   }
   os << "} {";  // tfm list
   for (intP  ip = tfmIndBegin;  ip != tfmIndEnd;  ip++)
   {
      os << tfmFFs[*ip].filename << ' ';
   }
   os << "} {";  // pk list
   for (intP  ip = pkIndBegin;  ip != pkIndEnd;  ip++)
   {
      os << pkFFs[*ip].filename << ' ';
   }
   os << "} }";
   
   delete [] ind;
   delete [] pkFFs;
   delete [] tfmFFs;
   return(ok);
} // matchFonts


#if defined(MAIN)

////////////////////////////////////////////////////////////////////////
int
main(int argc, char** argv)
{
   string mf;
   int  nTfm, nPk, nal;
   cin >> mf >> nTfm >> nPk >> nal;
   string*  s = new string[nal];
   const char**   ps = new const char*[nal];
   for (int  i = 0;  i < nal;  i++)
   {
      cin >> s[i];
      ps[i] = s[i].c_str();
   }

   bool  ok = matchFonts(cout, nTfm, nPk, nal, ps);

   cout << endl;

   return(!ok);
}
#endif
