// $Id: texpk.c,v 1.16 1997/07/11 05:06:00 yotam Exp $
// Class TeXpk - implementation

#include <iostream.h>
#include <fstream.h>
#include <BitSet.h>
#include "bmout.h"
#include "texpk.h"
#include "fintval.h"
#include "rect.h"
#include "verbose.h"
#include "debug.h"

#define VerboseMessage(vflag, el) _VerboseMessage(TeXpk::vflag, el)


extern void
bitmapOutput(ostream &os, BitSet** bitmap, US32 width, US32 height);

// for real implementation hiding
class _TeXpkImplement
{
private:
   enum {MaxChars = 256};
   class Char
   {
   public:
      _TeXpkImplement*  parent; // expensive but convinient for verbose

      US32       cc;  // convinient - although mostly the index
      US32       tfmWidth;
      Int32      dx;
      Int32      dy;
      US32       bitmapWidth;
      US32       bitmapHeight;
      Int32      hoff;
      Int32      voff;
      BitSet**   bitmap; // # bitmapHeight,  may repeat pointers!

      US32       nRects;
      Rect*      rectList;

      Char();
      Char(const Char &, _TeXpkImplement *);
      ~Char();

      int load(_TeXpkImplement* pparent,
               BifStream       &pkfile,
               US8             flag,
               int             verboseLoad);
      void clear();
   }; // Char


   int            _active;
   char*           _fileName; 

   // From preamble
   US8            _id;       // id=89 is currently supported
   char*          _comment;
   US32           _ds;
   US32           _cs;
   US32           _hppp;
   US32           _vppp;

   US16           _nChars;  // But only upto 256 supported.
   Char*          _charData[MaxChars];

   US8            _charFirst;
   US8            _charLast;
   US32           _maxOffset[TeXpk::NDir];

   void           calcOffsets();

public:

   US32           verbFlags;
   ostream*       verbStream;

   _TeXpkImplement();
   _TeXpkImplement(const _TeXpkImplement &);
   ~_TeXpkImplement();

   _TeXpkImplement &  operator = (const _TeXpkImplement &);

   void copyDynPart(const _TeXpkImplement &);

   void load(BifStream &pkfile, const char* fn);
   void clear();

   // retrieval members
   int           active()   { return(_active);};
   const char *  fileName() { return(_fileName);}; 
   US8           id()       { return(_id);};
   const char *  comment()  { return(_comment);};
   US32          ds()       { return(_ds);};
   US32          cs()       { return(_cs);};
   US32          hppp()     { return(_hppp);};
   US32          vppp()     { return(_vppp);};
   US16          nChars()   { return(_nChars);};
   const Char * const *  
                 charData() { return(_charData);};
   
   US8            charFirst() const {return _charFirst;}
   US8            charLast()  const {return _charLast;}
   US32           maxOffset(TeXpk::Direction di) const {return _maxOffset[di];}

   friend class  _TeXpkImplement::Char; 
}; // _TeXpkImplement


class PixelRun
{
public:
   const US32   bitmapWidth,
                bitmapHeight;
   US32         currRow,
                currCol,
                nextCol,
                runCount,
                repeatCount,
                currPixel;
   BitSet**       bitmap;
   Rect*         rectList;
   US32*         pnRects;

   PixelRun(
            US32       _bitmapWidth,
            US32       _bitmapHeight,
            US32       _currRow,
            US32       _currCol,
            US32       _nextCol,
            US32       _runCount,
            US32       _repeatCount,
            US32       _currPixel,
            BitSet **  _bitmap,
            Rect   *   _rectList,
            US32*       _pnRects
           ) :
           bitmapWidth(_bitmapWidth),
           bitmapHeight(_bitmapHeight),
           currRow(_currRow),
           currCol(_currCol),
           nextCol(_nextCol),
           runCount(_runCount),
           repeatCount(_repeatCount),
           currPixel(_currPixel),
           bitmap(_bitmap),
           rectList(_rectList),
           pnRects(_pnRects)
           {}
}; // PixelRun

////////////////////////////////////////////////////////////////////////
static US32
min(US32 n0, US32 n1)
{
   return(n0 < n1 ? n0 : n1);
} // min


////////////////////////////////////////////////////////////////////////
static char *
newStrCpy(char *& newStr, const char* str)
{
   int  rc = 1;
   int  l = strlen(str);
   newStr = new char[l + 1];
   rc = (newStr != 0);
   if (rc)
   {
      (void)strcpy(newStr, str);
   }
   return(newStr);
} // newStrCpy


////////////////////////////////////////////////////////////////////////
static void
charArrayRefDelete(char * &rpchar, US32 verbFlags, ostream* verbStream)
{
   if (rpchar)
   {
      delete [] rpchar;   VerboseDeleteA(rpchar);
      rpchar = 0;
   }
} // charArrayRefDelete


////////////////////////////////////////////////////////////////////////
_TeXpkImplement::Char::Char() :
   parent(0),
   bitmapHeight(0),
   bitmap(0),
   nRects(0),
   rectList(0)
{
   // DHERENL;
} // Char


////////////////////////////////////////////////////////////////////////
void
_TeXpkImplement::Char::clear()
{
   int     i;
   BitSet*  pbs, *pbsLast; // please support :

   US32  verbFlags = 0;
   ostream* verbStream = &cerr;
   if (parent != 0)
   {
      verbFlags = parent->verbFlags;
      verbStream = parent->verbStream;
   }

   DASSERT((bitmapHeight == 0) == (bitmap == 0));
   DASSERT((bitmapHeight == 0) == (rectList == 0));
   if (bitmap) // non empty
   {
      // The bitmap array may have consecutive elements
      // pointing to the  s a m e  BitSet pointer.
      for (pbsLast = 0, i = bitmapHeight;  i--;)
      {
         pbs = bitmap[i];
         if (pbs != pbsLast)
         {
            pbsLast = pbs;
            delete pbs;      VerboseDelete(pbs);
         }
      }
      delete [] bitmap;      VerboseDeleteA(bitmap);
      bitmap = 0;
      bitmapHeight = 0;
   }
   if (rectList)
   {
      delete [] rectList;    VerboseDeleteA(rectList);
      rectList = 0;
      nRects = 0;
   }
} // Char::clear


////////////////////////////////////////////////////////////////////////
_TeXpkImplement::Char::~Char()
{
   clear();
} // ~Char


////////////////////////////////////////////////////////////////////////
_TeXpkImplement::_TeXpkImplement() :
   _active(0),
   _fileName(0),
   _comment(0),
   verbFlags(0),
   verbStream(&cerr)
{
   DHERENL;
   for (int i = MaxChars;  i--;  _charData[i] = 0);
   for (int j = 4;  j--;  _maxOffset[j] = 0);
} // _TeXpkImplement


////////////////////////////////////////////////////////////////////////
void
_TeXpkImplement::clear()
{
   if (verbFlags & TeXpk::memory)
   {
      (*verbStream) << "TeXpk: clear" << endl;
   }
   // Yes, it could be optimized with link-list
   for (int i = MaxChars;  i--;)
   {
      if (_charData[i])
      {
         delete _charData[i];
         _charData[i] = 0;
      }
   }
   charArrayRefDelete(_fileName, verbFlags, verbStream);
   charArrayRefDelete(_comment,  verbFlags, verbStream);
   _active = 0;
} // clear


////////////////////////////////////////////////////////////////////////
_TeXpkImplement::~_TeXpkImplement()
{
   DHERENL;
   DASSERT((_active == 0) == (_fileName == 0));
   clear();
} // ~_TeXpkImplement


////////////////////////////////////////////////////////////////////////
static int
bitmapCheckSize(
   US32 *     prasterSize,
   US32       bitmapWidth,
   US32       bitmapHeight,
   ostream *  verbStream,
   US32       verbFlags
)
{
   int  rc = 1;

   if ((bitmapWidth == 0) || (bitmapHeight == 0))
   {
      *prasterSize = 0;
   }
   else if (bitmapWidth - 1 < ULONG_MAX / bitmapHeight)
   {
      *prasterSize = bitmapWidth * bitmapHeight;
   }
   else
   {
      Error("bitmap [" << dec << bitmapWidth << ',' << bitmapHeight <<
            "] is too large");
      rc = 0;
   }
   return(rc);
} // bitmapHeight


////////////////////////////////////////////////////////////////////////
// Inspired from GFtoPK
static US32
pk_packed_num(BifStream &pkfile, US32 dyn_f, US32 i)
{
   US32  j;

   DASSERT(i < 14); // starting nybble not a repeat count
   if (i == 0)
   {
      while ((j = pkfile.get4bits()) == 0)
      {
         i++;
      }
      for (i++;  i--;  j = (j << 4) | (US32)pkfile.get4bits());
      i = (j + 16*(13 - dyn_f) + dyn_f) - 15;
   }
   else if (i > dyn_f)
   {
      i = 16*(i - dyn_f - 1) + pkfile.get4bits() + dyn_f + 1;
   }
   return(i);
} // pk_packed_num


////////////////////////////////////////////////////////////////////////
static int
loadBitmap(BifStream & pkfile, PixelRun &pr)
{
   Rect*  pRect = &pr.rectList[0];
   for (pr.currRow = 0;  pr.currRow < pr.bitmapHeight;  pr.currRow++)
   {
      US8  black = 2; // undefined,
      for (pr.currCol = 0;  pr.currCol <= pr.bitmapWidth;  pr.currCol++)
      {
         // we fake a white right padded column
         US8   pixel = (pr.currCol < pr.bitmapWidth
                        ? pkfile.getByteBits(1)
                        : 0);
         DASSERT(pixel <= 1);
         if (black != pixel)
         {
            if (pixel)
            {            // initialize rectangle
               pRect->x = pr.currCol;
               pRect->y = pr.currRow;
               pRect->h = 1;
            }
            else if (black == 1)
            {            // add rectangle
               pRect->w = pr.currCol - pRect->x;
               pr.bitmap[pr.currRow]->set(pRect->x, pr.currCol - 1);
               pRect = &pr.rectList[++(*pr.pnRects)];
            }
            black = pixel;
         }
      }
   }
   return(1);
} // loadBitmapitmap


////////////////////////////////////////////////////////////////////////
static int
blackRun(PixelRun & pr, ostream * verbStream, US32 verbFlags)
{
   int   ok = 1;
   US32  lastCol = pr.nextCol - 1;
   US32  endCol = min(lastCol, pr.bitmapWidth - 1);
   pr.bitmap[pr.currRow]->set(pr.currCol, endCol);
   US32  &nRects = *pr.pnRects;
   pr.rectList[nRects++] = Rect(pr.currCol, pr.currRow,
                             endCol - pr.currCol + 1,
                             pr.repeatCount + 1);
   if (pr.nextCol >= pr.bitmapWidth)
   {
      pr.nextCol -= pr.bitmapWidth;
      pr.currRow += (pr.repeatCount + 1);
      pr.repeatCount = 0;
      US32  nFullRows = pr.nextCol / pr.bitmapWidth;

      // possible overflow was checked by nPixels

      lastCol %= pr.bitmapWidth;
      if (nFullRows)
      {
         pr.rectList[nRects++] = Rect(0, pr.currRow,
                                         pr.bitmapWidth, nFullRows);
         BitSet*  pfull = pr.bitmap[pr.currRow++];
         pfull->set();
         for (US32 r = nFullRows - 1; r--;)
         {
            delete pr.bitmap[pr.currRow];
            VerboseDelete(pr.bitmap[pr.currRow]);
            pr.bitmap[pr.currRow++] = pfull;
         }
      }
      if (lastCol != pr.bitmapWidth - 1)
      {
         pr.bitmap[pr.currRow]->set(0, lastCol);
         pr.rectList[nRects++] = Rect(0, pr.currRow, lastCol + 1, 1);
      }
   }
   return(ok);
} // blackRun


////////////////////////////////////////////////////////////////////////
static int
whiteRun(PixelRun & pr, ostream * verbStream, US32  verbFlags)
{
   int   ok = 1;
   if (pr.nextCol >= pr.bitmapWidth)
   {
      pr.currRow += (pr.repeatCount + 1);
      pr.repeatCount = 0;
      pr.nextCol -= pr.bitmapWidth;
      US32  nEmptyRows = pr.nextCol / pr.bitmapWidth;
      if (nEmptyRows)
      {
         BitSet*  pempty = pr.bitmap[pr.currRow++];
         for (nEmptyRows--; ok && nEmptyRows--;)
         {
            // possible overflow was checked by nPixels
            delete pr.bitmap[pr.currRow];
            VerboseDelete(pr.bitmap[pr.currRow]);
            pr.bitmap[pr.currRow++] = pempty;
         }
      }
   }
   return(ok);
} // whiteRun


////////////////////////////////////////////////////////////////////////
static void
reallocRects(Rect *& rectList, int nRects, ostream* verbStream, US32 verbFlags)
{
   Rect*  newRectList = 0;
   if (nRects > 0)
   {
      newRectList = new Rect[nRects];  VerboseNewA(newRectList);
      if (newRectList != 0)
      {
         for (int i = nRects;  i--;)
         {
            newRectList[i] = rectList[i];
         }
      }
   }
   if (rectList != 0)
   {
      delete [] rectList;                   VerboseDeleteA(rectList);
   }
   rectList = newRectList;
} // reallocRects


////////////////////////////////////////////////////////////////////////
int
_TeXpkImplement::Char::load(
   _TeXpkImplement*  pparent,
   BifStream &      pkfile,
   US8              flag,
   int              verboseLoad)

{
   // For the Error macro to work
   parent = pparent;
   US32      verbFlags  = pparent->verbFlags;
   ostream*  verbStream = pparent->verbStream;

#define VerboseLoad(el)  \
   if (verboseLoad) { (*verbStream) << el << endl; }

   US32       pl, rasterNbytes, rasterSize;
   int        ok = 1;
   US8        dyn_f     = (flag >> 4) & 0xf;  DASSERT(dyn_f != 0xf);
   int        bitmapped = (dyn_f == 14);
   int        blackMode = ((flag & 8) != 0);  // Is first run count black?
   US8        low3bits  = flag & 7;
   US32       low2bits  = low3bits & 3;
   streampos  ccPos, endRasterPos;

   dy = 0;
   if (low3bits <= 3)
   {  // Short form
      pl           = (low2bits << 8) | (US32)pkfile.get1ubyte();
      cc           = pkfile.get1ubyte();   ccPos = pkfile.tellg();
      tfmWidth     = pkfile.get3ubyte();
      dx           = pkfile.get1ubyte();
      bitmapWidth  = pkfile.get1ubyte();
      bitmapHeight = pkfile.get1ubyte();
      hoff         = pkfile.get1byte();
      voff         = pkfile.get1byte();
      rasterNbytes = pl - (3+1+1+1+1+1);
   }
   else if (low3bits != 7)
   {  // Extended short form
      pl           = (low2bits << 16) | (US32)pkfile.get2ubyte();
      cc           = pkfile.get1ubyte();   ccPos = pkfile.tellg();
      tfmWidth     = pkfile.get3ubyte();
      dx           = pkfile.get2ubyte();
      bitmapWidth  = pkfile.get2ubyte();
      bitmapHeight = pkfile.get2ubyte();
      hoff         = pkfile.get2byte();
      voff         = pkfile.get2byte();
      rasterNbytes = pl - (3+2+2+2+2+2);
   }
   else
   {  // Long form
      pl           = pkfile.get4ubyte();
      cc           = pkfile.get4ubyte();   ccPos = pkfile.tellg();
      tfmWidth     = pkfile.get4ubyte();
      dx           = pkfile.get4ubyte();
      dy           = pkfile.get4ubyte();
      bitmapWidth  = pkfile.get4ubyte();
      bitmapHeight = pkfile.get4ubyte();
      hoff         = pkfile.get4byte();
      voff         = pkfile.get4byte();
      rasterNbytes = pl - (3+4+4+4+4+4);
      if (cc >= 256)
      {
         Error("Unsupported character code " << cc << " >= 256");
         ok = 0;
      }
   }
   ok = ok && bitmapCheckSize(&rasterSize, bitmapWidth, bitmapHeight,
                              verbStream, verbFlags);
   if (ok)
   {
      VerboseLoad(hex <<
        "Load Character cc=" << (US16)cc << ", flag=" << (US16)flag << dec <<
                                                     ", pl=" << pl << endl <<
        "  tfm=" << tfmWidth << ", dx=" << dx << ", dy= " << dy <<
        ", bitmap[" << bitmapWidth << ',' << bitmapHeight << "], " <<
                                      "hoff= " << hoff << ", voff= " << voff);

      if (rasterSize)
      {
         bitmap   = new BitSet *[bitmapHeight];   VerboseNewA(bitmap);
         // rather generous bound for rects
         int maxRects = (bitmapped
                         ? ((bitmapWidth + 1)/2) * bitmapHeight
                         : 2 * (rasterNbytes + bitmapHeight));
         // For simplicity allocate BitSet for each row, repeats 'll be deleted
         rectList = new Rect[maxRects];                 VerboseNewA(rectList);
         ok = (bitmap != 0 && rectList != 0);
      }
      else
      {
         bitmap = 0;  rectList = 0;
      }
      for (US32 i = (bitmap != 0 ? bitmapHeight : 0);  i--;)
      {
          bitmap[i] = new BitSet; // Initialized as empty
          VerboseNew(bitmap[i]);
          if (bitmap[i] == 0)
          {
             ok = 0;
          }
      }
      if (!ok)
      {
         ok = 0;
         Error("Failed to allocate bitmap or rects" << endl <<
               "for height " << bitmapHeight << " and packet length " << pl);
      }
   }
   if (ok)
   {
      // now load the raseter data
      pkfile.bitscanFlush();
      PixelRun  pixelRun(bitmapWidth, bitmapHeight, 0, 0, 0,
                         0, 0, 0,
                         bitmap, rectList, &nRects);
      if (bitmapped)
      {
         VerboseLoad("Loading bitmap for cc=0x" << hex << cc);
         ok = loadBitmap(pkfile, pixelRun);
      }
      else
      {
         US32  nPixels;
         for (nPixels = 0;  ok && nPixels < rasterSize;)
         {
            US8  nybble = pkfile.get4bits();
            if (nybble >= 14)
            { // repeat count
               pixelRun.repeatCount = (
                  (nybble == 15)
                  ? 1
                  : pk_packed_num(pkfile, dyn_f, pkfile.get4bits()));
               nPixels += pixelRun.repeatCount * bitmapWidth;;
               if (nPixels <= rasterSize)
               {
                  DASSERT(pixelRun.currRow + pixelRun.repeatCount <
                          bitmapHeight);
                  for (US32  r = 1;  r <= pixelRun.repeatCount;  r++)
                  {
                     delete bitmap[pixelRun.currRow + r];
                     VerboseDelete(bitmap[pixelRun.currRow + r]);
                     bitmap[pixelRun.currRow + r] = bitmap[pixelRun.currRow];
                  }
                  if (nRects && (rectList[nRects - 1].y == pixelRun.currRow))
                  {  // fix 1st rectangle on this row
                     DASSERT(rectList[nRects -1].h == 1);
                     rectList[nRects -1].h = pixelRun.repeatCount + 1;
                  }
               }
               else
               {
                  ok = 0;
                  Error("Repeat count " << pixelRun.repeatCount <<
                        " bitmap row overflow");
               }
            }
            else
            {
               static const int (* const WBrun[2])(PixelRun &, ostream *, US32)
                                                 = {whiteRun, blackRun};
               pixelRun.runCount = pk_packed_num(pkfile, dyn_f, nybble);
               pixelRun.nextCol = pixelRun.currCol + pixelRun.runCount;
               nPixels += pixelRun.runCount;
               if (nPixels <= rasterSize)
               {

                  (*WBrun[blackMode])(pixelRun, verbStream, verbFlags);
                  pixelRun.currCol = pixelRun.nextCol % bitmapWidth;
                  blackMode = !blackMode;
               }
               else
               {
                  ok = 0;
                  Error("Run count " << pixelRun.runCount <<
                        " bitmap row overflow");
               }
            }
         }
         if (ok && ((pixelRun.currCol != 0) ||
                    (pixelRun.currRow != bitmapHeight) ||
                     nPixels != rasterSize))
         {
            ok = 0;
            Error("Raster data not synchronized");
         }
      }
      if (ok)
      {
         endRasterPos = pkfile.tellg();
         if ((streampos)(ccPos + pl) != endRasterPos)
         {
            ok = 0;
            Error("cc=0x" << hex << cc << dec <<
                  " (begin-postion=" << ccPos << ") + (pl=" << pl <<
                  ") != (end-postion=" << endRasterPos << ")");
         }
         if (ok)
         {
            reallocRects(rectList, nRects, verbStream, verbFlags);
         }
      }
      if (0 && ok && verboseLoad)
      {
         const BitSet * const * cBitmap = bitmap;
         bitmapOutput(*verbStream, cBitmap, bitmapWidth, bitmapHeight);
         rectsOutput(*verbStream, rectList, nRects);
      }

   }
   return(ok);
} // TeXpkImplement::Char::load


////////////////////////////////////////////////////////////////////////
static void
maxby(US32 &var, Int32 val)
{
   if ((Int32)var < val)
   {
      var = val;
   }
} // maxby


////////////////////////////////////////////////////////////////////////
void
_TeXpkImplement::calcOffsets()
{
   int  i;
   for (i = 4;  i--;  _maxOffset[i] = 0);
   for (i = 256, _charLast = 0;  i--;)
   {
      US8  c = (US8)i;
      Char*  pc = _charData[c];
      if (pc != 0)
      {
         Int32   w = pc->bitmapWidth;
         Int32   h = pc->bitmapHeight;
         Int32   hoff = pc->hoff;
         Int32   voff = pc->voff;
         maxby(_maxOffset[TeXpk::Left], hoff);
         maxby(_maxOffset[TeXpk::Right], w - hoff);
         maxby(_maxOffset[TeXpk::Up], voff);
         maxby(_maxOffset[TeXpk::Down], h - voff);
         if (_charLast == 0)  _charLast = c;
         _charFirst = c;         
      }
   }

} // calcOffsets


////////////////////////////////////////////////////////////////////////
void
_TeXpkImplement::load(BifStream &pkfile, const char* fn)
{
   enum pkOpcode
   {  // original naming as in GFtoPK weaved document
      pk_xxx1 = 0xf0,
      pk_xxx2,
      pk_xxx3,
      pk_xxx4,
      pk_yyy,
      pk_post,
      pk_no_op,
      pk_pre
   };

   bool  ok = true;
   if (newStrCpy(_fileName, fn) == 0)
   {
      ok = false;
      Error("Fail to copy file name " << fn);
   }
   US8   code;
   bool  preSeen = false;
   bool  postSeen = false;
   bool  anyCharYet = false;

   while (ok && !postSeen)
   {
      code = pkfile.get1ubyte();
      ok = (preSeen != (code == pk_pre));  // pk_pre iff the beginning
      if (!ok)
      {
         Error("Preamble not in the beginning");
      }
      else
      {
         if (code >= 0xf0)
         {
            switch (code)
            {
               case pk_xxx1:
                  pkfile.seekg( pkfile.get1ubyte(), ios::cur );
                  break;
               case pk_xxx2:
                  pkfile.seekg( pkfile.get2ubyte(), ios::cur );
                  break;
               case pk_xxx3:
                  pkfile.seekg( pkfile.get3ubyte(), ios::cur );
                  break;
               case pk_xxx4:
                  pkfile.seekg( pkfile.get4ubyte(), ios::cur );
                  break;
               case pk_yyy:
                  pkfile.seekg( 4, ios::cur );
                  break;
               case pk_post:
                  postSeen = true;
                  break;
               case pk_no_op:
                  break;
               case pk_pre:
                  preSeen = true;
                  _id = pkfile.get1ubyte();
                  if (_id != 89)
                  {
                     ok = 0;
                     Error("PK-id=" << _id << "!=89");
                  }
                  {
                     unsigned  commLen = pkfile.get1ubyte();
                     _comment = new char[commLen + 1];   VerboseNewA(_comment);
                     if (_comment == 0)
                     {
                        ok = 0;
                        Error("PK-comment memory failed");
                     }
                     else
                     {
                        pkfile.read(_comment, commLen);
                        _comment[commLen] = '\0';
                     }
                  }
                  _ds = pkfile.get4ubyte();
                  _cs = pkfile.get4ubyte();
                  _hppp = pkfile.get4ubyte();
                  _vppp = pkfile.get4ubyte();
                  VerbosePreamble("Preamble: " << endl <<
                   "  id=" << (int)_id <<
                   ", comment=<" << _comment << '>' << endl <<
                   "ds=" << _ds << ", cs=" << _cs <<
                   ", hppp=" << _hppp << ", vppp=" << _vppp);
                  break;
               default: Error("Bad PK opcode" << hex);
            }
         }
         else // character data
         {
            Char*  pChar = new(Char);   VerboseNew(pChar);
            if (pChar == 0)
            {
               ok = 0;
               Error("Failed to allocate memory for Character");
            }
            else
            {
               int verboseLoad = (verbFlags & TeXpk::allChars) ||
                             ((verbFlags & TeXpk::firstChar) && !anyCharYet);
               ok = pChar->load(this, pkfile, code, verboseLoad);
               anyCharYet = true;
               DASSERT(!ok || pChar->cc < MaxChars);
               if (ok)
               {
                  _charData[pChar->cc] = pChar;
               }
               else
               {
                  delete pChar;   VerboseDelete(pChar);
               }
            }
         }
      }
   }
   _active = ok;
   if (ok)
   {
      calcOffsets();
   }
   else
   {
      clear(); // also sets:  _active = 0
   }
} // _TeXpkImplement::load


////////////////////////////////////////////////////////////////////////
static int
newCopyBitmap(
   BitSet ** &                     toBitmap,
   const BitSet * const * const &  fromBitmap,
   US32                            bitmapHeight,
   ostream *                       verbStream,
   US32                            verbFlags
)
{
   int  ok = 1;
   toBitmap   = new BitSet *[bitmapHeight];   VerboseNewA(toBitmap);
   if (toBitmap == 0)
   {
      ok = 0;
      Error("No memory for bitmap copying");
   }
   else
   {
      const BitSet* pxLast = 0;
      BitSet*       pLast = 0;
      for (int i = bitmapHeight;  i--;)
      {
         if (pxLast != fromBitmap[i])
         {
            pxLast = fromBitmap[i];
            toBitmap[i] = pLast = new BitSet(*fromBitmap[i]);
                                                            VerboseNew(pLast);
            if (pLast == 0)
            {
               ok = 0;
               Error("No memory for bitmap-row copying");
            }
         }
         else
         {
            toBitmap[i] = pLast;  // same consecutive rows
         }
      }
   }
   return(ok);
} // newCopyBitmap


////////////////////////////////////////////////////////////////////////
_TeXpkImplement::Char::Char(const _TeXpkImplement::Char &x, _TeXpkImplement* p)
:
   parent(p),

   cc(x.cc),
   tfmWidth(x.tfmWidth),
   dx(x.dx),
   dy(x.dy),
   bitmapWidth(x.bitmapWidth),
   bitmapHeight(x.bitmapHeight),
   hoff(x.hoff),
   voff(x.voff)
{
   ostream*  verbStream = p->verbStream;
   US32     verbFlags   = p->verbFlags;
   int  ok = newCopyBitmap(bitmap, x.bitmap, bitmapHeight,
                           verbStream, verbFlags);
   rectList = new Rect[nRects];   VerboseNewA(rectList);
   if (rectList == 0)
   {
      ok = 0;
      Error("No memory for rects in copy constructor");
   }
   else
   {
      for (int i = nRects;  i--;)
      {
         rectList[i] = x.rectList[i];
      }
   }
   nRects = x.nRects;
   if (!ok)
   {
      p->_active = 0;
   }
} // _TeXpkImplement::Char(const TeXpk::Char &)


////////////////////////////////////////////////////////////////////////
void
_TeXpkImplement::copyDynPart(const _TeXpkImplement &x)
{
   bool  ok = true;

   if (newStrCpy(_fileName, x._fileName) == 0)
   {
      ok = false;
      Error("No memory for file name");
   }                                               VerboseNewA(_fileName);
   if (!newStrCpy(_comment, x._comment) == 0)
   {
      ok = false;
      Error("No memory for comment");
   }                                               VerboseNewA(_comment);
   for (int i = _nChars;  i--;)
   {
      _charData[i] = 0;
      if (x._charData[i])
      {
         _charData[i] = new Char(*x._charData[i], this);
                       VerboseNew(_charData[i]);
         if (_charData[i] == 0 || _active == 0)
         {
            ok = false;
            Error("Failed to initialize char data");
         }
      }
   }
   for (int k = 4;  k--;  _maxOffset[k] = x._maxOffset[k]);
   if (!ok)
   {
      clear();
   }
} // _TeXpkImplement::copyDynPart(const _TeXpkImplement &)


////////////////////////////////////////////////////////////////////////
_TeXpkImplement::_TeXpkImplement(const _TeXpkImplement &x)
:
   _active(x._active),
   _id(x._id),
   _ds(x._ds),
   _cs(x._cs),
   _hppp(x._hppp),
   _vppp(x._vppp),
   _nChars(x._nChars),
   verbFlags(x.verbFlags),
   verbStream(x.verbStream)
{
   if (_active)
   {
      copyDynPart(x);
   }
   _charFirst = x._charFirst;
   _charLast  = x._charLast;
   for (int k = 4;  k--;  _maxOffset[k] = x._maxOffset[k]);
} // _TeXpkImplement::_TeXpkImplement(const _TeXpkImplement &)


////////////////////////////////////////////////////////////////////////
_TeXpkImplement &
_TeXpkImplement::operator = (const _TeXpkImplement &x)
{
   clear();

   _active = x._active;
   verbStream = x.verbStream;
   verbFlags = x.verbFlags;

   if (_active)
   {
      _id        = x._id;
      _ds        = x._ds;
      _cs        = x._cs;
      _hppp      = x._hppp;
      _vppp      = x._vppp;
      _nChars    = x._nChars;
      _charFirst = x._charFirst;
      _charLast  = x._charLast;
      for (int k = 4;  k--;  _maxOffset[k] = x._maxOffset[k]);

      copyDynPart(x);
   }

   return(*this);
} // _TeXpkImplement &  operator = (const _TeXpkImplement &);



////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
//                     TeXpk  -  public members                       //
//                                                                    //


////////////////////////////////////////////////////////////////////////
TeXpk::TeXpk() :
   data(0)
{
} // TeXpk


////////////////////////////////////////////////////////////////////////
TeXpk::TeXpk(const char* fn, US32 verboseFlag, ostream& verbStream) :
   data(0)
{
   open(fn, verboseFlag, verbStream);
} // TeXpk


////////////////////////////////////////////////////////////////////////
TeXpk::TeXpk(const TeXpk &x) : data(0)
{
   if (x.data)
   {
      ostream*  verbStream = x.data->verbStream;
      US32      verbFlags  = x.data->verbFlags;
      data = new _TeXpkImplement(*x.data);    VerboseNew(data);
      if (data == 0)
      {
         Error("No memory for copy constructor");
      }
   }
} // TeXpk::TeXpk(const TeXpk &)


////////////////////////////////////////////////////////////////////////
TeXpk::~TeXpk()
{
   if (data)
   {
      US32    verbFlags = data->verbFlags;
      ostream* verbStream = data->verbStream;
      delete data;   VerboseDelete(data);
   }
} // ~TeXpk


////////////////////////////////////////////////////////////////////////
TeXpk &
TeXpk::operator = (const TeXpk &rhs)
{
   if (this != &rhs)
   {
      if (data)
      {
         *data = *rhs.data;
      }
   }
   return(*this);
} //TeXpk::operator =


////////////////////////////////////////////////////////////////////////
void
TeXpk::open(const char* fn, US32 verbFlags, ostream& rVerbStream)
{
   ostream* verbStream = &rVerbStream;
   int  ok = 1;
   if (data)
   {
      data->clear();
   }
   else
   {
      data = new _TeXpkImplement;  VerboseNew(data);
      if (data == 0)
      {
         ok = 0;
         Error("Failed to allocate data");
      }
   }
   BifStream pkfile(fn);        VerboseFclopen("Opening " << fn);
   if (!pkfile)
   {
      ok = 0;
      Error("Failed to open " << fn);
   }
   if (ok)
   {                            VerboseFclopen(fn << " opened");
      data->verbFlags = verbFlags;
      data->verbStream = verbStream;
      data->load(pkfile, fn);
      VerboseFclopen(fn << " closed");
   }
} // TeXpk::open


////////////////////////////////////////////////////////////////////////
int
TeXpk::active() const
{
   int  rc = (data ? data->active() : 0);
   return(rc);
} // TeXpk::active


////////////////////////////////////////////////////////////////////////
US8
TeXpk::id() const
{
   US8  rc = (data ? data->id() : 0);
   return(rc);
} // TeXpk::id


////////////////////////////////////////////////////////////////////////
const char*
TeXpk::comment() const
{
   const char*  rc = (data ? data->comment() : 0);
   return(rc);
} // TeXpk::comment


////////////////////////////////////////////////////////////////////////
US32
TeXpk::designSize() const
{
   US32  rc = (data ? data->ds() : 0);
   return(rc);
} // TeXpk::designSize


////////////////////////////////////////////////////////////////////////
US32
TeXpk::checkSum() const
{
   US32  rc = (data ? data->cs() : 0);
   return(rc);
} // TeXpk::checkSum


////////////////////////////////////////////////////////////////////////
US32
TeXpk::hppp() const
{
   US32  rc = (data ? data->hppp() : 0);
   return(rc);
} // TeXpk::hppp


////////////////////////////////////////////////////////////////////////
US32
TeXpk::vppp() const
{
   US32  rc = (data ? data->vppp() : 0);
   return(rc);
} // TeXpk::vppp


////////////////////////////////////////////////////////////////////////
US16
TeXpk::nChars() const
{
   US16  rc = (data ? data->nChars() : 0);
   return(rc);
} // TeXpk::nChars


////////////////////////////////////////////////////////////////////////
int
TeXpk::hasChar(US8 c) const
{
   int  rc = (data && (c < 256) ? (data->charData()[c] != 0) : 0);
   return(rc);
} // TeXpk::hasChar


////////////////////////////////////////////////////////////////////////
US32
TeXpk::tfmWidth(US8 c) const
{
   const _TeXpkImplement::Char*  pc;
   US32  rc = (data && (c < 256) && ((pc = data->charData()[c]))
               ? pc->tfmWidth
               : 0);
   return(rc);
} // TeXpk::tfmWidth


////////////////////////////////////////////////////////////////////////
Int32
TeXpk::dx(US8 c) const
{
   const _TeXpkImplement::Char*  pc;
   Int32  rc = (data && (c < 256) && ((pc = data->charData()[c])) 
                ? pc->dx 
                : 0);
   return(rc);
} // TeXpk::dx


////////////////////////////////////////////////////////////////////////
Int32
TeXpk::dy(US8 c) const
{
   const _TeXpkImplement::Char*  pc;
   Int32  rc = (data && (c < 256) && ((pc = data->charData()[c])) 
                ? pc->dy 
                : 0);
   return(rc);
} // TeXpk::dy


////////////////////////////////////////////////////////////////////////
US32
TeXpk::bitmapWidth(US8 c) const
{
   const _TeXpkImplement::Char*  pc;
   US32  rc = (data && (c < 256) && ((pc = data->charData()[c]))
               ? pc->bitmapWidth
               : 0);
   return(rc);
} // TeXpk::bitmapWidth


////////////////////////////////////////////////////////////////////////
US32
TeXpk::bitmapHeight(US8 c)  const
{
   const _TeXpkImplement::Char*  pc;
   US32  rc = (data && (c < 256) && ((pc = data->charData()[c]))
               ? pc->bitmapHeight
               : 0);
   return(rc);
} // TeXpk::bitmapHeight


////////////////////////////////////////////////////////////////////////
const BitSet * const *
TeXpk::bitmapBitSetArray(US8 c)  const
{
   const _TeXpkImplement::Char*  pc;
   const BitSet * const * rc = 
              (data && (c < 256) && ((pc = data->charData()[c]))
               ? pc->bitmap
               : 0);
   return(rc);
} // TeXpk::bitmapBitSetArray


////////////////////////////////////////////////////////////////////////
Int32
TeXpk::hoff(US8 c) const
{
   const _TeXpkImplement::Char*  pc;
   Int32  rc = (data && (c < 256) && ((pc = data->charData()[c])) 
                ? pc->hoff 
                : 0);
   return(rc);
} // TeXpk::hoff


////////////////////////////////////////////////////////////////////////
Int32
TeXpk::voff(US8 c) const
{
   const _TeXpkImplement::Char*  pc;
   Int32  rc = (data && (c < 256) && ((pc = data->charData()[c])) 
                ? pc->voff 
                : 0);
   return(rc);
} // TeXpk::voff


////////////////////////////////////////////////////////////////////////
US8
TeXpk::CharPixel(US8 c, US32 x, US32 y) const
{
   DASSERT(data && c <= 256 && data->charData()[c]);
   const _TeXpkImplement::Char*  pc = data->charData()[c];
   DASSERT(y < pc->bitmapHeight);
   US8  rc = pc->bitmap[y]->test(x);
   return(rc);
} // TeXpk::CharPixel


////////////////////////////////////////////////////////////////////////
US32
TeXpk::nRects(US8 c) const
{
   DASSERT(data && c <= 256 && data->charData()[c]);
   const _TeXpkImplement::Char*  pc = data->charData()[c];
   US32  rc = pc->nRects;
   return(rc);
} // TeXpk::nRects


////////////////////////////////////////////////////////////////////////
void
TeXpk::rect(US32 &x, US32 &y, US32 &w, US32 &h, US8 c, US32 rectInd) const
{
   DASSERT(data && c <= 256 && data->charData()[c]);
   const _TeXpkImplement::Char*  pc = data->charData()[c];
   const Rect*                   pr = &pc->rectList[rectInd];
   x = pr->x;
   y = pr->y;
   w = pr->w;
   h = pr->h;
} // TeXpk::rect


////////////////////////////////////////////////////////////////////////
US32
TeXpk::maxOffset(Direction di) const
{
   return(data != 0 ? data ->maxOffset(di) : 0);
} // TeXpk::maxOffset


////////////////////////////////////////////////////////////////////////
US8
TeXpk::charFirst() const
{
   return(data != 0 ? data->charFirst() : 0);
} // TeXpk::charFirst


////////////////////////////////////////////////////////////////////////
US8
TeXpk::charLast() const
{
   return(data != 0 ? data->charLast() : 0);
} // TeXpk::charLast


#if TEST
#include <stdlib.h>
int
main(int argc, char** argv)
{
   if (argc != 3)
   {
      cerr << "Usage:" << endl << argv[0] << "<pkfile> <veboseFlags>" << endl;
      return 1;
   }
   US32  verbFlags = 0;
   verbFlags = atol(argv[2]);
   TeXpk  texpk;
   texpk  = TeXpk(argv[1], verbFlags);
   return 0;
} // main
#endif
